import io
from flask import Flask, json, jsonify, redirect, render_template, send_file, send_from_directory, request, url_for, flash, Response
import logging
import yaml
import os
import csv
import csv
import qddate

from ..classify.processor import RulesProcessor
from metacrafter.classify.stats import Analyzer

DEFAULT_METACRAFTER_CONFIGFILE  ='.metacrafter'
DEFAULT_RULEPATH = ['rules',]
MANAGE_PREFIX = ''
DEFAULT_LIMIT = 1000

def prepare(processor):
    rulepath = []
    if os.path.exists(DEFAULT_METACRAFTER_CONFIGFILE):
        f = open(DEFAULT_METACRAFTER_CONFIGFILE, 'r', encoding='utf8')
        config = yaml.load(f, Loader=yaml.FullLoader)
        f.close()
        if config:
            if 'rulepath' in config.keys():
                rulepath = config['rulepath']
    else:
        rulepath = DEFAULT_RULEPATH
    for rp in rulepath:
        processor.import_rules_path(rp, recursive=True)

def scan_data():
    output = {}
    format = request.args.get('format', default='short', type=str)
    skipempty = request.args.get('skipempty', default=0, type=int)
    langs = request.args.get('langs', default=None, type=str)
    contexts = request.args.get('contexts', default=None, type=str)
    langs = langs.split('.') if langs is not None else None
    contexts = contexts.split('.') if contexts is not None else None
    try:
        processor = RulesProcessor()
        prepare(processor)
        items = json.loads(request.data)
        analyzer = Analyzer()
        datastats = analyzer.analyze(fromfile=None, itemlist=items,
                                     options={'delimiter': ',', 'format_in': None, 'zipfile': None})
        headers = ['key', 'ftype', 'is_dictkey', 'is_uniq', 'n_uniq', 'share_uniq', 'minlen', 'maxlen', 'avglen']
        datastats_dict = {}
        for row in datastats:
            datastats_dict[row[0]] = {}
            for n in range(0, len(headers)):
                datastats_dict[row[0]][headers[n]] = row[n]

        dparser = qddate.DateParser(patterns=qddate.patterns.PATTERNS_EN + qddate.patterns.PATTERNS_RU)
        results = processor.match_dict(items, datastats=datastats_dict, confidence=5, dateparser=dparser,
                                            parse_dates=True, filter_langs=langs, filter_contexts=contexts, limit=DEFAULT_LIMIT)

        if skipempty:
            output = []
            for o in results.results:
                if len(o.matches) != 0:
                    output.append(o.asdict())
        else:
            output = results.asdict()

    except Exception as ex:
        output = {'error' : 'Exception occured', 'message' : str(ex)}
    return jsonify(output)


def add_api_rules(app, prefix=MANAGE_PREFIX):
    """Adds API related rules to the"""
    app.add_url_rule(prefix + '/api/v1/scan_data', 'classify', scan_data, methods=['POST'],)



import chardet

def dict_generator(indict, pre=None):
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in list(indict.items()):
            if key == "_id":
                continue
            if isinstance(value, dict):
                #                print 'dgen', value, key, pre
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    if isinstance(v, dict):
                        #                print 'dgen', value, key, pre
                        for d in dict_generator(v, pre + [key]):
                            yield d
#                    for d in dict_generator(v, [key] + pre):
#                        yield d
            else:
                yield pre + [key, value]
    else:
        yield indict

def headers(data, limit=1000):
    n = 0
    keys = []
    for item in data:
        n += 1
        if n > limit: break
        dk = dict_generator(item)
        for i in dk:
            k = ".".join(i[:-1])
            if k not in keys:
                keys.append(k)
    return keys


def get_dict_value(d, keys):
    out = []
    if d is None:
        return out
#    keys = key.split('.')
    if len(keys) == 1:
        if type(d) == type({}):
            if keys[0] in d.keys():
                out.append(d[keys[0]])
        else:
            for r in d:
                if r and keys[0] in r.keys():
                    out.append(r[keys[0]])
#        return out
    else:
        if type(d) == type({}):
            if keys[0] in d.keys():
                out.extend(get_dict_value(d[keys[0]], keys[1:]))
        else:
            for r in d:
                if keys[0] in r.keys():
                    out.extend(get_dict_value(r[keys[0]], keys[1:]))
    return out


def dict_to_columns(data):
    columns = {}
    for row in data:
        dk = dict_generator(row)
        for i in dk:
            k = ".".join(i[:-1])
            if k in columns.keys():
                columns[k].append(i[-1])
            else:
                columns[k] = [i[-1],]
    return columns

def string_to_charrange(s):
    chars = {}
    for ch in s:
        v = chars.get(ch, 0)
        chars[ch] = v + 1
    return chars

def string_array_to_charrange(sarr):
    chars = {}
    for s in sarr:
        for ch in s:
            v = chars.get(ch, 0)
            chars[ch] = v + 1
    return chars


def detect_encoding(filename, limit=1000000):
    f = open(filename, 'rb')
    chunk = f.read(limit)
    f.close()
    detected = chardet.detect(chunk)
    return detected

def detect_delimiter(filename, encoding='utf8'):
    f = open(filename, 'r', encoding=encoding)
    line = f.readline()
    f.close()
    dict1 = {',': line.count(','), ';': line.count(';'), '\t': line.count('\t'), '|' : line.count('|')}
    delimiter = max(dict1, key=dict1.get)
    return delimiter


#!/usr/bin/env python
# -*- coding: utf8 -*-
import os
import logging
import csv
import qddate.patterns
import qddate
import click
import orjson
from tabulate import tabulate
from metacrafter.classify.processor import RulesProcessor
from metacrafter.classify.stats import Analyzer
from metacrafter.classify.utils import detect_delimiter, detect_encoding
import yaml
import bson
import pandas
import json

SUPPORTED_FILE_TYPES = ['jsonl', 'bson', 'csv', 'json']
BINARY_DATA_FORMATS = ['bson', 'parquet']

DEFAULT_METACRAFTER_CONFIGFILE = '.metacrafter'
DEFAULT_RULEPATH = ['rules',]

class CrafterCmd(object):

    def __init__(self):
        self.processor = RulesProcessor()
        self.prepare()
        pass


    def run_classifier_server(self):
        logging.info('Run server with classifier API')
        from metacrafter.server.manager import run_server
        self.prepare()
        run_server()


    def prepare(self):
        rulepath = []
        if os.path.exists(DEFAULT_METACRAFTER_CONFIGFILE):
            f = open(DEFAULT_METACRAFTER_CONFIGFILE, 'r', encoding='utf8')
            config = yaml.load(f, Loader=yaml.FullLoader)
            f.close()
            if config:
                if 'rulepath' in config.keys():
                    rulepath = config['rulepath']
        else:
            rulepath = DEFAULT_RULEPATH
        for rp in rulepath:
            self.processor.import_rules_path(rp, recursive=True)
        self.dparser = qddate.DateParser(patterns=qddate.patterns.PATTERNS_EN + qddate.patterns.PATTERNS_RU)

    def rules_load(self):
        self.processor.dumpStats()

    def _write_results(self, prepared, results, filename, dformat, output):
        if output:
            if output:
                f = open(output, 'w', encoding='utf8')
                out = []
                f.write(json.dumps({'table': filename, 'fields' : results}, indent=4, sort_keys=True))
                f.close()
                print('Output written to %s' % (output))
        elif len(prepared) > 0:
            outres = []
            if dformat == 'short':
                for r in prepared:
                    if len(r[3]) > 0:
                        outres.append(r)
                headers = ['key', 'ftype', 'tags', 'matches']
            elif dformat == 'full':
                outres = prepared
                headers = ['key', 'ftype', 'tags', 'matches']
            if len(outres) > 0:
                print(tabulate(outres, headers=headers))

    def _write_db_results(self, db_results, dformat, output):
        out = []
        for table, data in db_results.items():
            prepared, results = data
            if output:
                out.append({'table': table, 'fields': results})
            else:
                outres = []
                if dformat == 'short':
                    for r in prepared:
                        if len(r[3]) > 0:
                            outres.append(r)
                    headers = ['key', 'ftype', 'tags', 'matches']
                elif dformat == 'full':
                    outres = prepared
                    headers = ['key', 'ftype', 'tags', 'matches']
                if len(outres) > 0:
                    print(tabulate(outres, headers=headers))
        if output:
            print('Output written to %s' % (output))
            f = open(output, 'w', encoding='utf8')
            f.write(json.dumps(out, indent=4, sort_keys=True))
            f.close()

    def scan_data(self, items, limit=1000, contexts=None, langs=None):
        # load rules since file is acceptable

        analyzer = Analyzer()
        datastats = analyzer.analyze(fromfile=None, itemlist=items, options={'delimiter' : ',', 'format_in': None, 'zipfile': None})
        headers = ['key', 'ftype', 'is_dictkey', 'is_uniq', 'n_uniq', 'share_uniq', 'minlen', 'maxlen', 'avglen', 'tags']
        datastats_dict = {}
        for row in datastats:
            datastats_dict[row[0]] = {}
            for n in range(0, len(headers)):
                datastats_dict[row[0]][headers[n]] = row[n]

        results = self.processor.match_dict(items, datastats=datastats_dict, confidence=5, dateparser=self.dparser, parse_dates=True, limit=limit, filter_contexts=contexts, filter_langs=langs)
        output = []
        outdata = []
        for res in results.results:
            matches = []
            for match in res.matches:
                s = "%s %0.2f" % (match.dataclass, match.confidence)
                if match.format is not None:
                    s += ' (%s)' % (match.format)
                matches.append(s)
            if res.field not in datastats_dict.keys():
                continue
            output.append([res.field, datastats_dict[res.field]['ftype'], ','.join(datastats_dict[res.field]['tags']), ','.join(matches)])
            record = res.asdict()
            record['tags'] = datastats_dict[res.field]['tags']
            record['ftype'] = datastats_dict[res.field]['ftype']
            outdata.append(record)
        return output, outdata

    def scan_file(self, filename, delimiter=',', limit=1000, encoding='utf8', contexts=None, langs=None, dformat='short', output=None):
        ext = filename.rsplit('.', 1)[-1]
        if ext not in BINARY_DATA_FORMATS:
            encoding_dec = detect_encoding(filename)
            if encoding_dec:
                encoding = encoding_dec['encoding']
        filetype = None
        if ext == 'csv':
            filetype = 'csv'
            detected_delimiter = detect_delimiter(filename, encoding)
            if detected_delimiter:
                delimiter = detected_delimiter
            items = []
            f = open(filename, 'r', encoding=encoding)
            reader = csv.DictReader(f, delimiter=delimiter)
            n = 0
            for row in reader:
                n += 1
                if n > limit:
                    break
                items.append(row)
            f.close()
        elif ext == 'jsonl':
            filetype = 'jsonl'
            f = open(filename, 'r', encoding=encoding)
            items = []
            n = 0
            for l in f:
                n += 1
                if n > limit:
                    break
                items.append(orjson.loads(l))
            f.close()
        elif ext == 'bson':
            filetype = 'bson'
            f = open(filename, 'rb')
            items = []
            n = 0
            for l in bson.decode_file_iter(f):
                n += 1
                if n > limit:
                    break
                items.append(l)
            f.close()
        elif ext == 'json':
            filetype = 'json'
            f = open(filename, 'r', encoding=encoding)
            items = []
            n = 0
            data = json.load(f)
            if not isinstance(data, list) and not isinstance(data[0], dict):
                print(
                    'Unsupported JSON file. It should be "array JSON" with list of objects. Please preprocess data to this format')
                return []
            for l in data:
                n += 1
                if n > limit:
                    break
                items.append(l)
            f.close()
        elif ext == 'parquet':
            filetype = 'parquet'
            tbl = pandas.read_parquet(filename)
            items = []
            n = 0
            ad = tbl.to_dict('records')
            for l in ad:
                n += 1
                if n > limit:
                    break
                items.append(l)

        else:
            print('Unsupported file type. Supported file types are CSV, JSON lines, BSON, Parquet, JSON. Empty results')
            return []

        print('Filetype idenfied as %s' % (filetype))
        print('Processing file %s' % (filename))

        prepared, results  = self.scan_data(items, limit, contexts, langs)
        self._write_results(prepared, results, filename, dformat, output)


    def scan_db(self, connectstr='sqlite:///test.db', schema=None, limit=1000, contexts=None, langs=None, dformat='short', output=None):
        """SQL alchemy way to scan any database"""
        from sqlalchemy import create_engine, inspect
        import sqlalchemy.exc
        dbtype = connectstr.split(':', 1)[0].lower()
        print('Connecting to %s' % (connectstr))
        dbe = create_engine(connectstr)
        inspector = inspect(dbe)
        db_schemas = inspector.get_schema_names()
        con = dbe.connect()
        db_results = {}
        for db_schema in db_schemas:
            if schema and schema != db_schema:
                continue
            print("Processing schema: %s" % schema)
            if dbtype == 'postgres':
                con.execute('SET search_path TO {schema}'.format(schema=schema))
            for table in inspector.get_table_names(schema=schema):
                print('- table %s' % (table))
                try:
                    query = 'SELECT * FROM %s LIMIT %d' % (table, limit)
                    queryres = con.execute(query)
                except sqlalchemy.exc.ProgrammingError as e:
                    print('Error processing table %s: %s' % (table, str(e)))
                items = [dict(u) for u in queryres.fetchall()]
                prepared, results = self.scan_data(items, limit, contexts, langs)
                db_results[table] = [prepared, results]
        self._write_db_results(db_results, dformat, output)

    def scan_mongodb(self, host='localhost', port=27017, dbname='test', username=None, password=None, limit=1000, contexts=None, langs=None, dformat='short', output=None):
        """Scan entire MongoDB database"""
        print('Connecting to %s %d' % (host, port))
        from pymongo import MongoClient
        client = MongoClient(host, port, username=username, password=password)
        db = client[dbname]
        tables = db.list_collection_names()
        db_results = {}
        for table in tables:
            print('- table %s' % (table))
            items = list(db[table].find().limit(limit))
            prepared, results = self.scan_data(items, limit, contexts, langs)
            db_results[table] = [prepared, results]
        self._write_db_results(db_results, dformat, output)


@click.group()
def cli1():
    pass


@cli1.command()
def server():
    """Starts API and web interface for data management"""
    acmd = CrafterCmd()
    acmd.run_classifier_server()


@click.group()
def cli2():
    pass

@cli2.command()
def rules():
    """Rules load and test"""
    acmd = CrafterCmd()
    acmd.rules_load()

@click.group()
def cli3():
    pass

@cli3.command()
@click.argument('filename')
@click.option('--delimiter', '-d', default=',', help='CSV delimiter')
@click.option('--limit', '-n', default='1000', help='Limit of records')
@click.option('--contexts', '-x', default=None, help='List of contexts to use. Comma separates')
@click.option('--langs', '-l', default=None, help='List of languages to use. Comma separated')
@click.option('--format', '-f', default='short', help='Output format: short, long')
@click.option('--output', '-o', default=None, help='Output JSON filename')
def scan_file(filename, delimiter, limit, contexts, langs, format, output):
    """Match file"""
    acmd = CrafterCmd()
    acmd.scan_file(filename, delimiter, int(limit), contexts, langs, dformat=format, output=output)

@click.group()
def cli4():
    pass

@cli4.command()
@click.option('--connstr', '-c', default=None, help='SQLAlchemy connection string')
@click.option('--schema', '-s', default=None, help='Database schema. For Postgres DBs')
@click.option('--limit', '-n', default='1000', help='Limit of records')
@click.option('--contexts', '-x', default=None, help='List of contexts to use. Comma separates')
@click.option('--langs', '-l', default=None, help='List of languages to use. Comma separated')
@click.option('--format', '-f', default='short', help='Output format: short, long')
@click.option('--output', '-o', default=None, help='Output JSON filename')
def scan_db(connstr, schema, limit, contexts, langs, format, output):
    """Scan database using SQL alchemy connection string"""
    acmd = CrafterCmd()
    acmd.scan_db(connstr, schema, limit=int(limit), contexts=contexts, langs=langs, dformat=format, output=output)

@click.group()
def cli5():
    pass

@cli5.command()
@click.option('--host', '-h', default='localhost', help='SQLAlchemy connection string')
@click.option('--port', '-p', default=27017, help='Database schema. For Postgres DBs')
@click.option('--dbname', '-d', default="test", help='Database name')
@click.option('--username', '-u', default=None, help='Username. Optional')
@click.option('--password', '-P', default=None, help='Password. Optional')
@click.option('--limit', '-n', default='1000', help='Limit of records')
@click.option('--contexts', '-x', default=None, help='List of contexts to use. Comma separates')
@click.option('--langs', '-l', default=None, help='List of languages to use. Comma separated')
@click.option('--format', '-f', default='short', help='Output format: short, long')
@click.option('--output', '-o', default=None, help='Output JSON filename')
def scan_mongodb(host, port, dbname, username, password, limit, contexts, langs, format, output):
    """Scan MongoDB database"""
    acmd = CrafterCmd()
    acmd.scan_mongodb(host, int(port), dbname, username, password, limit=int(limit), contexts=contexts, langs=langs, dformat=format, output=output)



cli = click.CommandCollection(sources=[cli1, cli2, cli3, cli4, cli5])



from flask import Flask, json, jsonify, redirect, render_template, send_file, send_from_directory, request, url_for, flash, Response

import logging

from .api import add_api_rules


MANAGE_PREFIX = ''
CLASSIFY_HOST = '127.0.0.1'
CLASSIFY_PORT = 1399
DEBUG=False

SECRET_KEY = 'change_this_a_very_unique_secret_key'



def run_server():
#    global app

    app = Flask("Datacrafter", static_url_path='/assets')
    app.config['SECRET_KEY'] = SECRET_KEY
    app.config['PROPAGATE_EXCEPTIONS'] = True

    add_api_rules(app)


    logging.getLogger().addHandler(logging.StreamHandler())
    logging.basicConfig(
        filename='metacrafter_server.log',
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        level=logging.DEBUG,
    )

    app.run(host=CLASSIFY_HOST, port=CLASSIFY_PORT, debug=DEBUG)



# This is purely the result of trial and error.

import sys
import codecs

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

import metacrafter.classify


class PyTest(TestCommand):
    # `$ python setup.py test' simply installs minimal requirements
    # and runs the tests with no fancy stuff like parallel execution.
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--doctest-modules', '--verbose',
            './classifier', './tests'
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        sys.exit(pytest.main(self.test_args))


tests_require = [
    # Pytest needs to come last.
    # https://bitbucket.org/pypa/setuptools/issue/196/
    'pytest',
    'mock',
]


install_requires = [
    'pyparsing',
    'phonenumbers',
    'qddate',
    'orjson',
    'pymongo',
    'PyYAML',
    'pandas',
    'sqlalchemy',
    'validators'
]


# Conditional dependencies:

# sdist
if 'bdist_wheel' not in sys.argv:
    try:
        # noinspection PyUnresolvedReferences
        import argparse
    except ImportError:
        install_requires.append('argparse>=1.2.1')


# bdist_wheel
extras_require = {
    # https://wheel.readthedocs.io/en/latest/#defining-conditional-dependencies
    'python_version == "3.0" or python_version == "3.1"': ['argparse>=1.2.1'],
}


def long_description():
    with codecs.open('README.rst', encoding='utf8') as f:
        return f.read()


setup(
    name='metacrafter',
    version=metacrafter.classify.__version__,
    description=metacrafter.classify.__doc__.strip(),
    long_description=metacrafter.classify.__doc__.strip(),
#    long_description_content_type='text/rst',
    url='https://github.com/apicrafter/metacrafter/',
    download_url='https://github.com/apicrafter/metacrafter/',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    author=metacrafter.classify.__author__,
    author_email='ivan@begtin.tech',
    license=metacrafter.classify.__license__,
    entry_points={
        'console_scripts': [
            'metacrafter = metacrafter.__main__:main'
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    tests_require=tests_require,
    cmdclass={'test': PyTest},
    zip_safe=False,
    keywords='json jsonl csv bson cli dataset metadata',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development',
        'Topic :: System :: Networking',
        'Topic :: Terminals',
        'Topic :: Text Processing',
        'Topic :: Utilities'
    ],
)



from datetime import datetime, date
import logging
import bson
import orjson
import csv
import zipfile
from qddate import DateParser

DEFAULT_DICT_SHARE = 10
SUPPORTED_FILE_TYPES = ['xls', 'xlsx', 'csv', 'xml', 'json', 'jsonl', 'yaml', 'tsv', 'sql', 'bson']

DEFAULT_EMPTY_VALUES = [None, "", "None", "NaN", "-"]


DEFAULT_OPTIONS = {'encoding' : 'utf8',
                   'delimiter' : ',',
                   'limit' : 1000,
                   'empty' : DEFAULT_EMPTY_VALUES
                   }



def get_file_type(filename):
    ext = filename.rsplit('.', 1)[-1].lower()
    if ext in SUPPORTED_FILE_TYPES:
        return ext
    return None

def get_option(options, name):
    """Returns value of the option"""
    if name in options.keys():
        return options[name]
    elif name in DEFAULT_OPTIONS.keys():
        return DEFAULT_OPTIONS[name]
    return None


def guess_int_size(i):
    if i < 255:
        return 'uint8'
    if i < 65535:
        return 'uint16'
    return 'uint32'

def guess_datatype(s, qd):
    """Guesses type of data by string provided"""
    attrs = {'base' : 'str'}
#    s = unicode(s)
    if s is None:
       return {'base' : 'empty'}
    elif isinstance(s, bool):
        return {'base': 'bool'}
    elif isinstance(s, int):
        return {'base' : 'int'}
    elif isinstance(s, float):
        return {'base' : 'float'}
    elif isinstance(s, datetime):
        return {'base': 'datetime'}
    elif isinstance(s, date):
        return {'base': 'date'}
    elif type(s) != type(''):
#        print((type(s)))
        return {'base' : 'typed'}
#    s = s.decode('utf8', 'ignore')
    if s.isdigit():
        if s[0] == 0:
            attrs = {'base' : 'numstr'}
        else:
            attrs = {'base' : 'int', 'subtype' : guess_int_size(int(s))}
    else:
        try:
            i = float(s)
            attrs = {'base' : 'float'}
            return attrs
        except ValueError:
            pass
        if qd:
            is_date = False
            res = qd.match(s)
            if res:
                attrs = {'base': 'date', 'pat': res['pattern']}
                is_date = True
            if not is_date:
                if len(s.strip()) == 0:
                    attrs = {'base' : 'empty'}
    return attrs

def dict_generator(indict, pre=None):
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in list(indict.items()):
            if key == "_id":
                continue
            if isinstance(value, dict):
                #                print 'dgen', value, key, pre
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    if isinstance(v, dict):
                        #                print 'dgen', value, key, pre
                        for d in dict_generator(v, pre + [key]):
                            yield d
#                    for d in dict_generator(v, [key] + pre):
#                        yield d
            else:
                yield pre + [key, value]
    else:
        yield indict

class Analyzer:
    def __init__(self, nodates=True):
        if nodates:
            self.qd = None
        else:
            self.qd = DateParser(generate=True)
        pass

    def analyze(self, fromfile=None, itemlist=None, options=DEFAULT_OPTIONS):
        """Analyzes JSON file and produces stats"""
        if fromfile == None and itemlist == None:
            return None

        if 'empty' not in options.keys():
            options['empty'] = DEFAULT_EMPTY_VALUES

        dictshare = get_option(options, 'dictshare')
        if dictshare and dictshare.isdigit():
            dictshare = int(dictshare)
        else:
            dictshare = DEFAULT_DICT_SHARE

        profile = {'version': 1.0}
        fielddata = {}
        fieldtypes = {}

        #    data = json.load(open(profile['filename']))
        count = 0
        nfields = 0

        if fromfile:
            f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
            if f_type not in ['jsonl', 'bson', 'csv']:
                print('Only JSON lines (.jsonl), .csv and .bson files supported now')
                return

            if options['zipfile']:
                z = zipfile.ZipFile(fromfile, mode='r')
                fnames = z.namelist()
                finfilename = fnames[0]
                if f_type == 'bson':
                    infile = z.open(fnames[0], 'rb')
                else:
                    infile = z.open(fnames[0], 'r')
            else:
                finfilename = fromfile
                if f_type == 'bson':
                    infile = open(fromfile, 'rb')
                else:
                    infile = open(fromfile, 'r', encoding=get_option(options, 'encoding'))

            # Identify item list
            itemlist = []

            if f_type == 'jsonl':
                for l in infile:
                    itemlist.append(orjson.loads(l))
            elif f_type == 'csv':
                delimiter = get_option(options, 'delimiter')
                reader = csv.DictReader(infile, delimiter=delimiter)
                for r in reader:
                    itemlist.append(r)
            elif f_type == 'bson':
                bson_iter = bson.decode_file_iter(infile)
                for r in bson_iter:
                    itemlist.append(r)


        # process data items one by one
        logging.debug('Start processing %s' % (fromfile))
        for item in itemlist:
            count += 1
            dk = dict_generator(item)
            if count % 1000 == 0: logging.debug('Processing %d records of %s' % (count, fromfile))
            for i in dk:
                #            print(i)
                k = '.'.join(i[:-1])
                if len(i) == 0: continue
                if i[0].isdigit(): continue
                if len(i[0]) == 1: continue
                v = i[-1]
                if k not in list(fielddata.keys()):
                    fielddata[k] = {'key': k, 'uniq': {}, 'n_uniq': 0, 'total': 0, 'share_uniq': 0.0, 'minlen' : None, 'maxlen' : 0, 'avglen' : 0, 'totallen' : 0}
                fd = fielddata[k]
                uniqval = fd['uniq'].get(v, 0)
                fd['uniq'][v] = uniqval + 1
                fd['total'] += 1
                if uniqval == 0:
                    fd['n_uniq'] += 1
                    fd['share_uniq'] = (fd['n_uniq'] * 100.0) / fd['total']
                fl = len(str(v))
                if fd['minlen'] is None:
                    fd['minlen'] = fl
                else:
                    fd['minlen'] = fl if fl < fd['minlen'] else fd['minlen']
                fd['maxlen'] = fl if fl > fd['maxlen'] else fd['maxlen']
                fd['totallen'] += fl
                fielddata[k] = fd
                if k not in list(fieldtypes.keys()):
                    fieldtypes[k] = {'key': k, 'types': {}}
                fd = fieldtypes[k]
                thetype = guess_datatype(v, self.qd)['base']
                uniqval = fd['types'].get(thetype, 0)
                fd['types'][thetype] = uniqval + 1
                fieldtypes[k] = fd
        #        print count
        for k, v in list(fielddata.items()):
            fielddata[k]['share_uniq'] = (v['n_uniq'] * 100.0) / v['total']
            fielddata[k]['avglen'] = v['totallen'] / v['total']
        profile['count'] = count
        profile['num_fields'] = nfields
        finfields = {}
        for k, v in list(fielddata.items()):
#            del v['uniq']
            fielddata[k] = v
        for fdk in list(fieldtypes.values()):
            fdt = list(fdk['types'].keys())
            if 'empty' in fdt:
                del fdk['types']['empty']
            if len(list(fdk['types'].keys())) != 1:
                ftype = 'str'
            else:
                ftype = list(fdk['types'].keys())[0]
            finfields[fdk['key']] = ftype

        dictkeys = []
        dicts = {}
#        print(profile)
        profile['fields'] = []
        for fd in list(fielddata.values()):
#            print(fd)
#            print(fd['key'])  # , fd['n_uniq'], fd['share_uniq'], fieldtypes[fd['key']]
            field = {'key': fd['key'], 'is_uniq': 0 if fd['share_uniq'] < 100 else 1}
            profile['fields'].append(field)
            if fd['share_uniq'] < dictshare:
                dictkeys.append(fd['key'])
                dicts[fd['key']] = {'items': fd['uniq'], 'count': fd['n_uniq'],
                                    'total' : sum(fd['uniq'].values()),
                                    'type': finfields[fd['key']]}  # TODO: Shouldn't be "str" by default
        #            for k, v in fd['uniq'].items():
        #                print fd['key'], k, v
        profile['dictkeys'] = dictkeys

        profile['debug'] = {'fieldtypes': fieldtypes.copy(), 'fielddata': fielddata}
        profile['fieldtypes'] = finfields
        table = []
        for fd in list(fielddata.values()):
            field = [fd['key'], ]
            field.append(finfields[fd['key']])
            field.append(True if fd['key'] in dictkeys else False)
            field.append(False if fd['share_uniq'] < 100 else True)
            field.append(fd['n_uniq'])
            field.append(fd['share_uniq'])
            field.append(fd['minlen'])
            field.append(fd['maxlen'])
            field.append(fd['avglen'])
            tags = []
            if fd['share_uniq'] == 100:
                tags.append('uniq')
            allempty = 0
            if fd['key'] in dicts.keys():
                for key, value in dicts[fd['key']]['items'].items():
                    if key in options['empty']:
                        allempty += value
                if allempty == dicts[fd['key']]['total']:
                    tags.append('empty')
                else:
                    tags.append('dict')
            field.append(tags)
            table.append(field)
        return table

    def print(self, table):
        from tabulate import tabulate
        headers = ['key', 'ftype', 'is_dictkey', 'is_uniq', 'n_uniq', 'share_uniq', 'minlen', 'maxlen', 'avglen', 'tags']
        print(tabulate(table, headers=headers))



import logging

from pyparsing import Literal, Optional, ParseException, White, Word, alphanums, alphas, hexnums, lineEnd, lineStart, nums, oneOf, CaselessLiteral, Group, OneOrMore, Combine, printables
import yaml
import glob
import importlib
from pathlib import Path

from .utils import headers, get_dict_value, dict_to_columns

DEFAULT_MAX_LEN = 100
DEFAULT_MIN_LEN = 3

RULE_TYPE_FIELD = 1
RULE_TYPE_DATA = 2

BASE_URL = 'https://meta.apicrafter.io/class/{dataclass}'


class TableScanResult:
    """Results of table scan classification"""
    def __init__(self, results=[]):
        self.results = results
        pass

    def add(self, result):
        """Add one more Columnt scan result"""
        self.results.append(result)

    def is_empty(self):
        """True if empty"""
        return len(self.results) == 0

    def asdict(self):
        res = []
        for m in self.results:
            res.append(m.asdict())
        return res


class ColumnMatchResult:
    """Results of the column classficiations """
    def __init__(self, field, matches=[]):
        self.field = field
        self.matches = matches
        pass

    def add(self, match):
        """Add one more rule result"""
        self.matches.append(match)

    def is_empty(self):
        """True if empty"""
        return len(self.matches) == 0

    def asdict(self):
        matches = []
        for m in self.matches:
            matches.append(m.asdict())
        return {'field' : self.field, 'matches' : matches}



class RuleResult:
    """Result match error"""
    def __init__(self, ruleid, dataclass, confidence, ruletype, piiclass=None, format=None):
        self.ruleid = ruleid
        self.dataclass = dataclass
        self.confidence = confidence
        self.ruletype = ruletype
        self.piiclass = piiclass
        self.format = format

    def is_pii(self):
        return self.piiclass is not None

    def class_url(self):
        return BASE_URL.format(dataclass=self.dataclass)

    def pii_url(self):
        return BASE_URL.format(dataclass=self.piiclass) if self.piiclass is not None else None

    def asdict(self):
        return {'ruleid' : self.ruleid, 'dataclass' : self.dataclass, 'confidence' : self.confidence, 'ruletype' : self.ruletype, 'format' : self.format, 'classurl' : self.class_url()}


class RulesProcessor:
    """Classification rules processor class"""
    def __init__(self, langs=None, contexts=None):
        self.preset_langs = langs
        self.preset_contexts = contexts
        self.reset_rules()
        pass

    def reset_rules(self):
        """Cleans up imported rules."""
        self.data_rules = []
        self.field_rules = []
        self.__rule_keys = []
        self.langs = {}
        self.contexts = {}


    def import_rules(self, filename):
        """Import rules from file"""
        f = open(filename, 'r', encoding='utf8')
        ruledata = yaml.load(f, Loader=yaml.FullLoader)
        f.close()

        # If group of rules context or lang not in allowed list, skip it
        if self.preset_langs and ruledata['lang'] not in self.preset_langs:
            return
        if self.preset_contexts and ruledata['context'] not in self.preset_contexts:
            return

        for rulekey in ruledata['rules'].keys():
            if rulekey in self.__rule_keys:
                continue
            else:
                self.__rule_keys.append(rulekey)
            rule = ruledata['rules'][rulekey]

            rule['imprecise'] = (int(rule['imprecise']) != 0) if 'imprecise' in rule.keys() else False
#            print(rulekey, rule['imprecise'])
            if rule['match'] == 'ppr':
                rule['compiled'] = lineStart + eval(rule['rule']) + lineEnd
            elif rule['match'] == 'func':
                module, funcname = rule['rule'].rsplit('.', 1)
                match_func = getattr(importlib.import_module(module), funcname)
                rule['compiled'] = match_func
            elif rule['match'] == 'text':
                keywords = rule['rule'].split(',')
                ruledata['rules'][rulekey]['compiled'] = lineStart + oneOf(keywords, caseless=True) + lineEnd
                ruledata['rules'][rulekey]['keywords'] = list(map(str.lower, keywords))
            if rule['match'] == 'text':
                rule['maxlen'] = len(max(keywords, key=len))
                rule['minlen'] = len(min(keywords, key=len))
            else:
                rule['maxlen'] = int(rule['maxlen']) if 'maxlen' in rule.keys() else DEFAULT_MAX_LEN
                rule['minlen'] = int(rule['minlen']) if 'minlen' in rule.keys() else DEFAULT_MIN_LEN
            if 'validator' in rule.keys():
                module, funcname = rule['validator'].rsplit('.', 1)
                match_func = getattr(importlib.import_module(module), funcname)
                rule['vfunc'] = match_func
            if 'fieldrule' in rule.keys() and 'fieldrulematch' in rule.keys():
                if rule['fieldrulematch'] == 'ppr':
                    rule['f_compiled'] = lineStart + eval(rule['fieldrule']) + lineEnd
                elif rule['fieldrulematch'] == 'text':
                    keywords = rule['fieldrule'].split(',')
                    ruledata['rules'][rulekey]['fieldkeywords'] = list(map(str.lower, keywords))
                    ruledata['rules'][rulekey]['f_compiled'] = lineStart + oneOf(keywords, caseless=True) + lineEnd
            rule['id'] = rulekey
            for key in ['context', 'lang']:
                rule[key] = ruledata[key]
            rule['group'] = ruledata['name']
            rule['group_desc'] = ruledata['description']
            if rule['type'] == 'field':
                self.field_rules.append(rule)
            elif rule['type'] == 'data':
                self.data_rules.append(rule)

            v = self.langs.get(rule['lang'], 0)
            self.langs[rule['lang']] = v + 1

            contexts = rule['context'].split('.')
            rule['context'] = contexts

            # Add more than one context
            if 'piikey' in rule.keys() and 'pii' not in rule['context']:
                rule['context'].append('pii')

            for context in contexts:
                v = self.contexts.get(context, 0)
                self.contexts[context] = v + 1


        logging.debug('Loaded rules from %s' % filename)


    def import_rules_path(self, pathname, recursive=True):
        """Import rules from path"""
        if not recursive:
            filenames = glob.glob(pathname + '/*.yaml')
            for filename in filenames:
                self.import_rules(filename)
        else:
            for path in Path(pathname).rglob('*.yaml'):
                self.import_rules(str(path))

    def dumpStats(self):
        """Dump statistics"""
        import qddate

        print('Rule types:')
        print('- field based rules %d' % (len(self.field_rules)))
        print('- data based rules %d' % (len(self.data_rules)))
        print('Context:')
        for key in sorted(self.contexts.keys()):
            print('- %s %d' % (key, self.contexts[key]))
        print('Language:')
        for key in sorted(self.langs.keys()):
            print('- %s %d' % (key, self.langs[key]))
        dparser = qddate.DateParser(patterns=qddate.patterns.PATTERNS_EN + qddate.patterns.PATTERNS_RU)
        print('Data/time patterns (qddate): %d' % (len(dparser.patterns)))

    def get_filtered_rules(self, ruletype, contexts=None, langs=None, ignore_imprecise=False):
        """Filters rules by ruletype, context and languages. Helps to filter rules before data matching"""
        rules = self.field_rules if ruletype == RULE_TYPE_FIELD else self.data_rules
        filtered = []
        if not contexts and not langs and not ignore_imprecise:
            return rules
        for rule in rules:
            in_context = False
            in_lang = False
            in_imprecise = False
            if not contexts:
                in_context = True
            else:
                for rule_context in rule['context']:
                    if rule_context in contexts:
                        in_context = True
                        break
            if not langs:
                in_lang = True
            elif rule['lang'] in langs:
                in_lang = True
            if ignore_imprecise and rule['imprecise']:
                in_imprecise = True
            if in_lang and in_context and not in_imprecise:
                filtered.append(rule)
#            else:
#                print('Rule %s removed' % (rule['key']))
#            print(rule['key'], in_lang, in_context, in_imprecise)
        return filtered

    def match_dict(self, data, fields=None, datastats=None, confidence=95, stop_on_match=False, parse_dates=True, dateparser=None, limit=1000, filter_contexts=None, filter_langs=None, except_empty=True, ignore_imprecise=True):
        """Matches python array of dicts (from JSON lines or BSON)"""
        results = TableScanResult()
        if not fields:
            fields = headers(data)
        field_rules = self.get_filtered_rules(RULE_TYPE_FIELD, filter_contexts, filter_langs, ignore_imprecise)
        data_rules = self.get_filtered_rules(RULE_TYPE_DATA, filter_contexts, filter_langs, ignore_imprecise)

        data_columns = dict_to_columns(data)
        nonstr = []
        if datastats:
            for field in datastats.keys():
                if datastats[field]['ftype'] != 'str':
                    nonstr.append(field)
        for field in fields:
            m_result = ColumnMatchResult(field=field, matches=[])
            shortfield = field.rsplit('.', 1)[-1].strip()
            # Match field name first
            for rule in field_rules:
                if rule['match'] == 'func':
                    res = rule['compiled'](shortfield)
                    if res:
                        m_result.add(RuleResult(ruleid=rule['id'], confidence=100, dataclass=rule['key'], ruletype='field'))
                        if stop_on_match:
                            break
                elif rule['match'] == 'ppr':
                    try:
                        res = rule['compiled'].parseString(shortfield)
                        m_result.add({'dataclass': rule['key'], 'confidence': 100, 'ruleid' : rule['id'], 'ruletype' : 'field'})
                        if stop_on_match:
                            break
                    except ParseException:
                        pass
                elif rule['match'] == 'text':
                    if shortfield.lower() in rule['keywords']:
                        m_result.add(RuleResult(ruleid=rule['id'], confidence=100, dataclass=rule['key'], ruletype='field'))
                        if stop_on_match:
                            break
            data = data_columns[field]
            slice = data[0:limit]
            min_len = 0
            max_len = 0
            if datastats:
                # Match boolean as bool type and ignore float fields right now #FIXME
                if field in datastats.keys():
                    if datastats[field]['ftype'] == 'bool':
                        m_result.add(RuleResult(ruleid='_int_fieldtype_boolean', confidence=100, dataclass='boolean', ruletype='fieldtype'))
                        results.add(m_result)
                        continue
                    elif datastats[field]['ftype'] == 'float':
                        results.add(m_result)
                        continue
                    elif datastats[field]['ftype'] == 'datetime':
                        m_result.add(RuleResult(ruleid='_int_fieldtype_datetime', confidence=100, dataclass='datetime', ruletype='fieldtype'))
                        results.add(m_result)
                        continue
                    elif datastats[field]['ftype'] == 'date':
                        m_result.add(RuleResult(ruleid='_int_fieldtype_date', confidence=100, dataclass='date', ruletype='fieldtype'))
                        results.add(m_result)
                        continue
                    min_len = datastats[field]['minlen']
                    max_len = datastats[field]['maxlen']
            if min_len == 0:
                try:
                    min_len = len(min(data, key=len))
                    max_len = len(max(data, key=len))
                except TypeError:
                    min_len = 4
                    max_len = 4

#            if field not in nonstr:

            if field:
                # Filtering rules by field max and min length of the data
                rules = []
                for rule in data_rules:
#                    print('Field %s %d %d, rule %s %d %d' % (field, min_len, max_len, rule['key'], rule['minlen'], rule['maxlen']))
                    if (min_len >= rule['minlen'] and min_len <= rule['maxlen']) or (rule['minlen'] >= min_len and rule['minlen'] <= max_len):
                        if 'f_compiled' in rule.keys():
                                if rule['fieldrulematch'] == 'ppr':
                                    try:
                                        res = rule['f_compiled'].parseString(shortfield)
                                        rules.append(rule)
                                    except ParseException:
                                        pass
                                elif rule['fieldrulematch'] == 'text':
                                    if shortfield.lower() in rule['fieldkeywords']:
                                        rules.append(rule)
                        else:
                            rules.append(rule)
#                print(field)
#                for rule in rules:
#                    print('- %s' %(rule['key']))
                for rule in rules:
                    success = 0
                    empty = 0
                    total = len(slice)
                    for value in slice:
                        if not value:
                            if except_empty:
                                empty += 1
                            continue
#                        if not isinstance(value, str): continue
                        slen = len(str(value))
                        if slen == 0:
                            if except_empty:
                                empty += 1
                            continue

                        if slen < rule['minlen'] or slen > rule['maxlen']: continue
                        if rule['match'] == 'func':
                            try:
                                res = rule['compiled'](str(value))
                                if res:
                                    success += 1
                            except KeyboardInterrupt:
                                pass
                        elif rule['match'] == 'ppr':
                            try:
                                res = rule['compiled'].parseString(str(value))
                                if 'vfunc' in rule.keys():
                                    isvalid = rule['vfunc'](str(value))
                                    if isvalid:
                                        success += 1
                                else:
                                    success += 1
                            except ParseException as e:
                                pass
                        elif rule['match'] == 'text':
                            if str(value).lower() in rule['keywords']:
                                success += 1
                    if except_empty:
                        subtotal = total - empty
                        if subtotal == 0:
                            result = 0
                        else:
                            result = (success * 100.0 / (total - empty))
                    else:
                        result = success * 100.0 / total
                    if result > confidence:
                        m_result.add(RuleResult(ruleid=rule['id'], confidence=result, dataclass=rule['key'], ruletype='data'))
                        if stop_on_match:
                            break

            if m_result.is_empty() and parse_dates and field not in nonstr:
                total = len(slice)
                success = 0
                empty = 0
                date_format = None
                for value in slice:
                    if value is None:
                        if except_empty:
                            empty += 1
                        continue
                    if len(str(value)) == 0:
                        if except_empty:
                            empty += 1
                        continue
                    if not isinstance(value, str): continue
                    res = dateparser.match(value, noyear=False)
                    if res:
                        success += 1
                        date_format = res['pattern']['key']

                if except_empty:
                    subtotal = total - empty
                    if subtotal == 0:
                        result = 0
                    else:
                        result = (success * 100.0 / (total - empty))
                else:
                    result = success * 100.0 / total
                if result > confidence:
                    m_result.add(
                        RuleResult(ruleid='qddate', confidence=result, dataclass='datetime', ruletype='data', format=date_format))
            results.add(m_result)
        return results
